#!/usr/bin/python
import pdb, re, copy, os.path, math, numpy
from toolLog import Log

def readLogicLine(line, patt=None):
    triples = []
    if patt is None:
        patt = ['^(?P<pred>[\w:.,;#=]+)\s*\(\s*(?P<NodeOutN>[\w:.,;#]+)\s*,\s*(?P<NodeInN>[\w:.,;#]+)\s*\)\s*$',
                '^(?P<pred>[\w]+)\s*\(\s*(?P<NodeOutN>[\w:.,;#]+)\s*\)\s*$']

    tmp = re.match(patt[0], line)
    if tmp is not None:
        triples.append( (tmp.group("NodeOutN"), tmp.group('pred'), tmp.group('NodeInN')) )
    else:
        tmp = re.match(patt[1], line)
        if tmp is not None:
            triples.append( (tmp.group("NodeOutN"), tmp.group('pred'), None) )
    return triples

def readBMGLine(line, patt=None):
    triples = []
    if patt is None:
        patt = ['^([A-Z]_)(?P<NodeOutN>[\w:.,;#]+)\s+([A-Z]_)(?P<NodeInN>[\w:.,;#]+)\s+(?P<Comm>.*)$', \
                '^# _attributes ([A-Z]_)(?P<NodeN>[\w:.,;#]+)\s*(?P<Comm>.*)$']

    tmp = re.match(patt[1], line)
    if tmp is not None:
        for att in tmp.group("Comm").split():
            parts = att.split("=", 1)
            if len(parts) == 2:
                triples.append( (tmp.group("NodeN"), parts[0], parts[1]) )
    else:
        tmp = re.match(patt[0], line)
        if tmp is not None:
            for att in tmp.group("Comm").split():
                triples.append( (tmp.group("NodeOutN"), att, tmp.group('NodeInN')) )
    return triples

def readTripleLine(line):
    return [line.split()]

class Graph:
    OUT=0; IN=1
    EG=0; ND=1
    COMP_EQ = 0
    COMP_LT = 1
    COMP_GT = -1
    singleton_label = "singleton"
    comp_match = "^Na[<>=]Nb$"
    greater_match = "^Na>Nb$"
    comp_labels = ['Na=Nb', 'Na<Nb', 'Na>Nb']
    log = Log()
    
    def typeE(elem):
        if type(elem) in [list, tuple] and len(elem) == 2:
            return Graph.EG
        else:
            return Graph.ND
    typeE = staticmethod(typeE)

    def isEG(elem):
        return type(elem) in [list, tuple] and len(elem) == 2
    isEG = staticmethod(isEG)

    def isComp(cons):
        return re.match(Graph.comp_match, cons[1]) is not None
    isComp = staticmethod(isComp)

    def isCompGreater(cons):
        return re.match(Graph.greater_match, cons[1]) is not None
    isCompGreater = staticmethod(isCompGreater)

    def isSingleton(cons):
        return cons[0] == Graph.singleton_label
    isSingleton = staticmethod(isSingleton)

    def onlyComp(atts):
        outcome = True
        for pred, val in atts:
            outcome &= Graph.isComp((pred, val))
        return outcome
    onlyComp = staticmethod(onlyComp)

    def hasSingleton(atts):
        outcome = False
        for pred, val in atts:
            outcome |= Graph.isSingleton((pred, val))
        return outcome
    hasSingleton = staticmethod(hasSingleton)

    def __init__(self):
        self.elems_pred = [{},{}]
        self.elems = [{}, {}]
        self.node_dict = None

    def singletonElem(self, elem):
        return self.elems[Graph.typeE(elem)].get(elem, {}).has_key(Graph.singleton_label)

    def project(self, nodeset):
        edges = self.elems[Graph.EG].keys()
        for edge in edges:
            if edge[0] not in nodeset or edge[1] not in nodeset:
                del self.elems[Graph.EG][edge]
        nodes = self.elems[Graph.ND].keys()
        for node in nodes:
            if node not in nodeset:
                del self.elems[Graph.ND][node]
        self.updateElemsPreds()

    def merge(self, other):
        for tE in [Graph.ND, Graph.EG]:
            for elem, atts in other.elems[tE].items():
                if self.elems[tE].has_key(elem):
                    self.elems[tE][elem].update(atts)
                else:
                    self.elems[tE][elem] = atts
        self.updateElemsPreds()

    def remove(self, other):
        for tE in [Graph.ND, Graph.EG]:
            for elem, atts in other.elems[tE].items():
                if self.elems[tE].has_key(elem):
                    for att, val in atts.items():
                        if self.elems[tE][elem].has_key(att) and self.elems[tE][elem][att] == val:
                            del self.elems[tE][elem][att]
                    if len(self.elems[tE][elem]) == 0:
                        del self.elems[tE][elem]
        for edge in self.elems[Graph.EG].keys():
            for endi in [0, 1]:
                if not self.elems[tE].has_key(elem):
                    self.elems[tE][elem] = {}
        self.updateElemsPreds()

    def getElem(self, elem):
        return self.elems[Graph.typeE(elem)].get(elem, {})

    def getComponents(self, check_comp=True):
        if check_comp:
            edges = []
            for edge, atts in self.elems[Graph.EG].items():
                if not Graph.onlyComp(atts.items()): ### use only connecting edges
                    edges.append(edge)
        else:
            edges = self.elems[Graph.EG].keys()
        nodes = self.elems[Graph.ND].keys()
        part_to_nodes = dict([(i, [n]) for (i,n) in enumerate(nodes)])
        nodes_to_part = dict([(n, i) for (i,n) in enumerate(nodes)])
        for edge in edges:
            if nodes_to_part[edge[0]] != nodes_to_part[edge[1]]:
                if nodes_to_part[edge[0]] < nodes_to_part[edge[1]]:
                    edge = [edge[1], edge[0]]
                part_from = nodes_to_part[edge[1]]
                part_to = nodes_to_part[edge[0]]
                for node in part_to_nodes[part_from]:
                    nodes_to_part[node] = part_to
                part_to_nodes[part_to].extend(part_to_nodes[part_from])
                del part_to_nodes[part_from]
        return part_to_nodes, nodes_to_part

    def hanging(self, queries=None):
        if queries is None:
            queries = set()
        cse = {}
        for se in set([tuple(sorted(e)) for e in self.elems[Graph.EG].keys()]):
            for side in [0,1]:
                if cse.has_key(se[side]):
                    cse[se[side]].append(se[1-side])
                else:
                    cse[se[side]] = [se[1-side]]
        hanging = []
        new_hanging = [node for node, neigh in cse.items() if node not in queries and len(neigh) == 1]
        while len(new_hanging) > 0:
            hanging.extend(new_hanging)
            for node in new_hanging:        
                cse[cse[node][0]].remove(node)
                del cse[node]
            new_hanging = [node for node, neigh in cse.items() if node not in queries and len(neigh) == 1]
        return set(hanging)
        
    def updateElems(self):
        self.elems = {Graph.EG: {}, Graph.ND: {}}
        for tE in [Graph.ND, Graph.EG]:
            for pred, entries in self.elems_pred[tE].items():
                if type(entries) == list:
                    for elem, val in entries:
                        if not self.elems[tE].has_key(elem):
                            self.elems[tE][elem] = {}
                        self.elems[tE][elem][pred] = val
                else:
                    for val, elems in entries.items():
                        for elem in elems:
                            if not self.elems[tE].has_key(elem):
                                self.elems[tE][elem] = {}
                            self.elems[tE][elem][pred] = val
                            
    def updateElemsPreds(self):
        self.node_dict = None
        self.elems_pred = {Graph.EG: {}, Graph.ND: {}}
        for tE in [Graph.ND, Graph.EG]:
            for elem, entries in self.elems[tE].items():
                for pred, val in entries.items():
                    if type(val) in [int, float]:
                        if not self.elems_pred[tE].has_key(pred):
                            self.elems_pred[tE][pred] = []
                        self.elems_pred[tE][pred].append((elem, val)) 
                    else:
                        if not self.elems_pred[tE].has_key(elem):
                            self.elems_pred[tE][pred]= {}
                        if not self.elems_pred[tE][pred].has_key(val):
                            self.elems_pred[tE][pred][val] = []
                        self.elems_pred[tE][pred][val].append(elem)

    def copy(self):
        tmp = Graph()
        tmp.elems_pred = copy.deepcopy(self.elems_pred)
        tmp.elems = copy.deepcopy(self.elems)
        return tmp

    def nodeId(self, node):
        if self.node_dict is None:
            self.node_dict = dict([(v,k) for k,v in enumerate(sorted(self.elems[Graph.ND].keys()))])
        try:
            return self.node_dict[node]
        except ValueError:
            return None
        
    def translateNode(self, node_from, node_to, update_preds=True):
        if self.elems[Graph.ND].has_key(node_from) and not self.elems[Graph.ND].has_key(node_to):
            old_edges = self.elems[Graph.EG].keys()
            for edge in old_edges:
                if node_from == edge[0]:
                    new_edge = (node_to, edge[1])
                    preds = self.elems[Graph.EG].pop(edge)
                    self.elems[Graph.EG][new_edge] = preds
                if node_from == edge[1]:
                    new_edge = (edge[0], node_to)
                    preds = self.elems[Graph.EG].pop(edge)
                    self.elems[Graph.EG][new_edge] = preds

            preds = self.elems[Graph.ND].pop(node_from)
            self.elems[Graph.ND][node_to] = preds
            if update_preds:
                self.updateElemsPreds()
            return True
        return False

    def translate(self, translation_dict):
        done = []
        for node_from, node_to in translation_dict.items():
            if self.translateNode(node_from, node_to, False):
                done.append(node_from)
        if len(done) > 0:
            self.updateElemsPreds()
        return done
                    
    def getTypesPred(self):
        tmp = [] # edges, nodes
        for tE in [Graph.EG, Graph.ND]:
            tmp.extend(self.elems_pred[tE].keys())
        return tmp

    def getConstraintsList(self):
        return self.getEdges(True)+self.getNodes(True)

    def nbConstraints(self, tE=None):
        if tE is not None:
            typesE = [tE]
        else:
            typesE = [Graph.ND, Graph.EG]
        counts = 0
        for tE in typesE:
            counts += sum([len(atts) for elem, atts in self.elems[tE].items()])
        return counts

    def getEdges(self, check_empty=False):
        if check_empty:
            return [e for e in self.elems[Graph.EG].items() if len(e[1]) > 0]
        else:
            return self.elems[Graph.EG].items()

    def getNodes(self, check_empty=False):
        if check_empty:
            return [e for e in self.elems[Graph.ND].items() if len(e[1]) > 0]
        else:
            return self.elems[Graph.ND].items()

    def nbEdges(self, check_empty=False):
        return len(self.getEdges())

    def nbNodes(self, check_empty=False):
        return len(self.getNodes())
        
    def edgesSet(self):
        return set(self.elems[Graph.EG].keys())

    def nodesSet(self):
        return set(self.elems[Graph.ND].keys())

    def neighborsSub(self, direction = 0, tmp_preds=None):
        ndict = dict([(node, set()) for node in self.nodesSet()])
        if tmp_preds is None:
            preds = self.elems_pred[Graph.EG].keys()
        else:
            preds = [k for k in self.elems_pred[Graph.EG].keys() if k in tmp_preds]

        if direction == -1:
            directions = [0,1]
        else:
            directions = [direction]
        for pred in preds:
            for val, edges_val in self.elems_pred[Graph.EG][pred].items():
                for edge in edges_val:
                    for direc in directions:
                        ndict[edge[direc]].add(edge[1-direc])
        return ndict

    def elemPreds(self, elem, preds=None):
        tE = Graph.typeE(elem)
        if preds is None:
            return self.elems[tE].get(elem, {}).items()
        else:
            return [(k,v) for (k,v) in self.elems[tE].get(elem, {}).items() if k in preds]

    def compareNodes(self, nodeA, nodeB, comp_pred=[]):
        comp_set = []
        if len(comp_pred) > 0:
            if self.elems[Graph.ND].has_key(nodeA) and self.elems[Graph.ND].has_key(nodeB):
                for pred in comp_pred:
                    if self.elems[Graph.ND][nodeA].has_key(pred) and self.elems[Graph.ND][nodeB].has_key(pred):
                        vS = self.elems[Graph.ND][nodeA][pred]
                        vE = self.elems[Graph.ND][nodeB][pred]
                        comp_set.append((pred, Graph.comp_labels[(vS < vE) - (vS > vE)]))
        return comp_set

    def elementMatch(self, matchElem, consts):
        tE = Graph.typeE(matchElem)
        match = True
        ci = 0
        while ci < len(consts) and match:
            pred, val = consts[ci]
            ci +=1
            if Graph.isSingleton((pred, val)):
                match &= matchElem == val
            elif Graph.isComp((pred, val)):
                match &= (len(self.filterCompare([matchElem], pred, val)) == 1)
            else:
                ### WARNING: FIRST TEST CAN'T BE MOVED OUT, IN CASE OF COMPARISON EDGES...
                match &= (self.elems[tE].has_key(matchElem) and self.elems[tE][matchElem].has_key(pred) and self.elems[tE][matchElem][pred]==val)
        return match

    def matchingElements(self, tE, consts):
        postproc = []
        tmp_map = None
        for (pred, val) in consts:
            if Graph.isComp((pred, val)):
                postproc.append((pred, val))
            else:
                tmp_c = set(self.getConstraintMatch(tE, pred, val))
                if tmp_map == None:
                    tmp_map = tmp_c
                else:
                    tmp_map &= tmp_c
        if tmp_map != None:
            tmp_map = list(tmp_map)
            for (pred, val) in postproc:
                tmp_map = self.filterCompare(tmp_map, pred, val)
        return tmp_map

    def filterCompare(self, tmp_map, pred, val):
        filtered_list = []
        tE = Graph.ND
        for (nodeA, nodeB) in tmp_map:
            if self.elems[tE][nodeA].has_key(pred) and self.elems[tE][nodeB].has_key(pred):
                vS = self.elems[tE][nodeA][pred]
                vE = self.elems[tE][nodeB][pred]
                if val == Graph.comp_labels[(vS < vE) - (vS > vE)] :
                    filtered_list.append((nodeA, nodeB))
        return filtered_list
            
    def getConstraintMatch(self, typeE, pred, val=None): #, filtered=False):
        #### This is not for comparisons!!
        if Graph.isSingleton((pred, val)):
            if self.elems[typeE].has_key(val):
                return [val]
            else:
                return []
        return self.elems_pred[typeE].get(pred, {}).get(val,[])

    def getConstrainedPath(self, path, pred_types):
        node_pred = []; edge_pred = []; comp_pred = []        
        for (k,v) in pred_types[Graph.ND].items():
            if v == "cat":
                node_pred.append(k)
            elif v == "num":
                comp_pred.append(k)
        edge_pred = pred_types[Graph.EG].keys()

        constraintsP = []
        elemi = -1
        for elemi in range(len(path)-1):
            constraintsP.append(self.elemPreds(path[elemi], node_pred))
            constraintsP.append(self.elemPreds((path[elemi], path[elemi+1]), edge_pred))
            constraintsP.append(self.elemPreds((path[elemi+1], path[elemi]), edge_pred))
            for elemj in range(elemi+1):
                constraintsP.append(self.compareNodes(path[elemj], path[elemi+1], comp_pred))
        constraintsP.append(self.elemPreds(path[elemi+1], node_pred))
        return constraintsP


    def parseTriples(self, triples, pred_types):
        for triple in triples:
            self.parseTriple(triple, pred_types)

    def parseTriple(self, triple, pred_types):            
        if triple is not None and len(triple) > 2:
            tE = None
            (x, pred, y) = triple[:3]
            if y is None or (pred_types[Graph.ND] is not None and pred in pred_types[Graph.ND]):
                if pred_types[Graph.ND] is not None and pred_types[Graph.ND][pred] == "num":
                    try:
                        y = float(y)
                    except ValueError:
                        return
                tE = Graph.ND
                elem = x
                val = y
                if y is None:
                    val = 'ND'
                self.parseElem(tE, elem, pred, val)
            else:
                val = 'EG'
                tmp = pred.split("=", 1)
                if len(tmp) > 1:
                    val = tmp[1]
                    pred = tmp[0]
                if ( pred_types[Graph.EG] is None and pred_types[Graph.ND] is None) \
                       or pred in pred_types[Graph.EG]:
                    if pred_types[Graph.EG] is not None \
                           and val != "EG" \
                           and pred_types[Graph.EG][pred] == "num":
                        val = float(val)
                    tE = Graph.EG
                    elem = (x,y)
                    self.parseElem(tE, elem, pred, val)

    def parseElem(self, tE, elem, pred, val):
        if tE == Graph.EG:
            for end_i in [0,1]:
                if not self.elems[Graph.ND].has_key(elem[end_i]):
                    self.elems[Graph.ND][elem[end_i]] = {}
            
        if self.elems[tE].has_key(elem):
            self.elems[tE][elem][pred] = val
        else:
            self.elems[tE][elem] = {pred: val}

        if self.elems_pred[tE].has_key(pred):
            if type(val) in [float, int]:
                self.elems_pred[tE][pred].append((elem, val))                        
            elif self.elems_pred[tE][pred].has_key(val):
                self.elems_pred[tE][pred][val].append(elem)
            else:
                self.elems_pred[tE][pred][val] = [elem]
        else:
            if type(val) in [float, int]:
                self.elems_pred[tE][pred]= [(elem,val)]
            else:
                self.elems_pred[tE][pred]= {val:[elem]}

    def listTriples(self):
        triples = []
        for x, pvs in self.elems[Graph.ND].items():
            for pred, val in pvs.items():
                triples.append((x, pred, val))
                
        for (x,y), preds in self.elems[Graph.EG].items():
            for pred, val in preds.items():
                if val != "EG":
                    triples.append((x, "%s=%s" % (pred, val), y))
                else:
                    triples.append((x, pred, y))
        return triples

    def writeTriples(self, fp):
        triples = []
        for x, pvs in self.elems[Graph.ND].items():
            for pred, val in pvs.items():
                fp.write("%s %s %s\n" % (x, pred, val))
                
        for (x,y), preds in self.elems[Graph.EG].items():
            for pred, val in preds.items():
                if val != "EG":
                    fp.write("%s %s=%s %s\n" % (x, pred, val, y))
                else:
                    fp.write("%s %s %s\n" % (x, pred, y))

    def writeBMG(self, fp, patt_replace=r'^', replacement='T_', add_att=None):
        for x, pvs in self.elems[Graph.ND].items():
            xf = re.sub(patt_replace, replacement, x)
            fp.write("# _attributes %s " % xf)
            if add_att is not None:
                fp.write(add_att)
            fp.write(" ".join(["%s=%s" % (pred, val) for pred, val in pvs.items()])+"\n")
                
        for (x,y), preds in self.elems[Graph.EG].items():
            xf = re.sub(patt_replace, replacement, x)
            yf = re.sub(patt_replace, replacement, y)
            for pred, val in preds.items():
                if val != "EG":
                    fp.write("%s %s %s=%s\n" % (xf, yf, pred, val))
                else:
                    fp.write("%s %s %s\n" % (xf, yf, pred))

    def listBMG(self, patt_replace='^', replacement='T_', add_att=None):
        lines = []
        for x, pvs in self.elems[Graph.ND].items():
            xf = re.sub(patt_replace, replacement, x)
            lines.append("# _attributes %s " % xf)
            if add_att is not None:
                lines[-1] += add_att
            lines[-1]+= " ".join(["%s=%s" % (pred, val) for pred, val in pvs.items()])
                
        for (x,y), preds in self.elems[Graph.EG].items():
            xf = re.sub(patt_replace, replacement, x)
            yf = re.sub(patt_replace, replacement, y)
            for pred, val in preds.items():
                if val != "EG":
                    lines.append("%s %s %s=%s" % (xf, yf, pred, val))
                else:
                    lines.append("%s %s %s" % (xf, yf, pred))
        return lines

    def writeLogic(self, fp, patt_replace=r'^', replacement='T_'):
        for x, pvs in self.elems[Graph.ND].items():
            xf = re.sub(patt_replace, replacement, x)
            for pred, val in pvs.items():
                fp.write("%s(%s,%s)\n" % (pred, xf, val))
                
        for (x,y), preds in self.elems[Graph.EG].items():
            xf = re.sub(patt_replace, replacement, x)
            yf = re.sub(patt_replace, replacement, y)
            for pred, val in preds.items():
                if val != "EG":
                    fp.write("%s=%s(%s,%s)\n" % (pred, val, xf, yf))
                else:
                    fp.write("%s(%s,%s)\n" % (pred, xf, yf))

    def listLogic(self, patt_replace=r'^', replacement='T_'):
        lines = []
        for x, pvs in self.elems[Graph.ND].items():
            xf = re.sub(patt_replace, replacement, x)
            for pred, val in pvs.items():
                lines.append("%s(%s,%s)\n" % (pred, xf, val))
                
        for (x,y), preds in self.elems[Graph.EG].items():
            xf = re.sub(patt_replace, replacement, x)
            yf = re.sub(patt_replace, replacement, y)
            for pred, val in preds.items():
                if val != "EG":
                    lines.append("%s=%s(%s,%s)\n" % (pred, val, xf, yf))
                else:
                    lines.append("%s(%s,%s)\n" % (pred, xf, yf))
        return lines

    def mapClauseStraight(self, constraints_list, query_nodes=None, prev_node_map=None, oi_check=True):
        Graph.log.printL(4, "Start mapping")
        Graph.log.printL(5, "Constraints:")
        for const in constraints_list:
            Graph.log.printL(5, "\t%s:%s" % const)
        expand = 1

        if prev_node_map is None:
            node_map = NodeMap()
        else:
            node_map = prev_node_map.copy()
        
        for next_elem, atts_elem in constraints_list:
            Graph.log.printL(5, "Now mapping %s" % str(next_elem))
            if expand <= 0:
                break

            if Graph.isEG(next_elem):
                #### MAP EDGE
                for endi in [0,1]:
                    if not node_map.hasNode(next_elem[endi]):
                        node_map.addNode(next_elem[endi])

                if node_map.sameFragment(next_elem):
                    expand = node_map.mapClosingEdge(next_elem, atts_elem.items(), self)
                else:
                    match_edges = self.matchingElements(Graph.EG, atts_elem.items())
                    expand, new_fragid, old_fragid = node_map.mapJoiningEdge(next_elem, match_edges, oi_check)

            else:
                #### MAP NODE
                expand = node_map.mapNode(next_elem, self.matchingElements(Graph.ND, atts_elem.items()))

        if expand > 0:
            Graph.log.printL(4, "Mapping alright %d..." % expand)
        else:
            Graph.log.printL(4, "Mapping interrupted %d..." % expand)
        if expand == 0: Warning("Went badly wrong...")
  
        details = {'status': expand}
        return node_map, details


    def mapClause(self, constraints_list, query_nodes=None, prev_node_map=None, oi_check=True, max_score=0):
        Graph.log.printL(4, "Start mapping (max_score=%d)" % max_score)
        Graph.log.printL(5, "Constraints:")
        for const in constraints_list:
            Graph.log.printL(5, "\t%s:%s" % const)
        expand = 1

        if prev_node_map is None:
            node_map = NodeMap()
        else:
            node_map = prev_node_map.copy()

        nodes = set()
        edges_data = {}
        for next_elem, atts_elem in constraints_list:
            if Graph.isEG(next_elem):
                edges_data[next_elem] = {'constraints': atts_elem.items() ,'match_edges': None}
                edges_data[next_elem]['comp'] = Graph.onlyComp(edges_data[next_elem]['constraints'])
                nodes.update(next_elem)
            else:
                #### MAP NODE ALREADY
                expand = node_map.mapNode(next_elem, self.matchingElements(Graph.ND, atts_elem.items()))

        for node in nodes - node_map.nodesSet():
            node_map.addNode(node)

        ### DONE WITH NODES, LET'S TAKE CARE OF EDGES NOW
        cards = []
        done_edges = []
        expand = 1
        frag_edges, closing_edges = node_map.edgesFragments(edges_data.keys())

        ### prepare connectivity infos
        heads_present = node_map.containedNodes(query_nodes)

        head_fragments = node_map.fragmentsNodes(heads_present)
        if len(head_fragments) == 1:
            cards.append((node_map.nbInstances(heads_present[0]), len(done_edges)))

        while expand > 0 and len(closing_edges) > 0:
            next_elem = closing_edges.pop()
            Graph.log.printL(5, "Now mapping %s, (%d) CLOSING" % (next_elem, expand))
            expand = node_map.mapClosingEdge(next_elem, edges_data[next_elem]['constraints'], self)
            done_edges.append(next_elem)
            if len(head_fragments) == 1 and node_map.sameFragment((next_elem[0], heads_present[0])):
                cards.append((node_map.nbInstances(heads_present[0]), len(done_edges)))

        close_id = len(done_edges)

        next_map = -1 ### just to get into the loop
        
        while next_map is not None and expand > 0:
            #### choose next edge to map
            next_map = None
            candidates_categories = [set(), set(), set()]
            for edge in set(edges_data.keys())-set(done_edges):
                if not edges_data[edge]['comp']:
                    inheads = 0
                    for endi in [0,1]:
                        if node_map.nodeFragment(edge[endi]) in head_fragments:
                            inheads += 1
                    candidates_categories[inheads].add(edge)

            while next_map is None and len(candidates_categories) > 0:
                edges_candidates = candidates_categories.pop()
                for edge in edges_candidates:
                    if edges_data[edge]['match_edges'] is None:
                        edges_data[edge]['match_edges'] = self.matchingElements(Graph.EG, edges_data[edge]['constraints'])

                    score = node_map.estimateCount(edge, edges_data[edge]['match_edges'])
                    if max_score > 0 and score > max_score:
                        Graph.log.printL(5, "TOOO HIGH SCORE BOTH %s, (%d)" % (edge, score))
                        score *= -1
                    if score > 0 and (next_map is None or next_map[1] > score):
                        next_map = (edge, score)

            if next_map is not None:
                next_elem = next_map[0]
                Graph.log.printL(5, "Now mapping %s, (%d) JOINING %d" % (next_elem, expand, next_map[1]))

                expand, new_fragid, old_fragid = node_map.mapJoiningEdge(next_elem, edges_data[next_elem]['match_edges'], oi_check)
                if expand > next_map[1]:
                    Graph.log.printL(2, "For %s reality is over estimate that's odd (%d vs. %d)..." % (next_map[0], expand, next_map[1]))
                frag_edges[new_fragid].remove(next_elem)

                ### update infos about head-connectedness, frags cardinalities, etc.
                done_edges.append(next_elem)

                head_fragments = node_map.fragmentsNodes(heads_present)
                if len(head_fragments) == 1 and node_map.sameFragment((next_elem[0], heads_present[0])):
                    cards.append((node_map.nbInstances(heads_present[0]), len(done_edges)))
                    
                ### Find closing edges and close...
                closing_edges = list(frag_edges[new_fragid] & frag_edges[old_fragid])
                frag_edges[new_fragid].update(frag_edges[old_fragid])
                while len(closing_edges) > 0 and expand > 0:
                    next_closing = closing_edges.pop()

                    ### next_elem has been remove from one side so it is not among closing edges
                    Graph.log.printL(5, "Now mapping %s, (%d) CLOSING" % (next_closing, expand))
                    expand = node_map.mapClosingEdge(next_closing, edges_data[next_closing]['constraints'], self)

                    done_edges.append(next_closing)

                    if len(head_fragments) == 1 and node_map.sameFragment((next_closing[0], heads_present[0])):
                        cards.append((node_map.nbInstances(heads_present[0]), len(done_edges)))

        reprocess = set(edges_data.keys())-set(done_edges)
        if len(reprocess)== 0 and expand > 0:
            Graph.log.printL(4, "Mapping alright %d..." % expand)
        else:
            expand *= -1
            Graph.log.printL(4, "Mapping interrupted %d..." % expand)
        if expand == 0: Warning("Went badly wrong...")

        details = {'status': expand, 'reprocess': reprocess, 'done': done_edges, 'cards': cards, 'close_id': close_id, 'heads_present': heads_present, 'head_fragments': head_fragments}
        return node_map, details


    def generateClauses(self, constraints_list, query_nodes=None, oi_check=True, max_score=0):
        Graph.log.printL(4, "Start mapping (max_score=%d)" % max_score)
        Graph.log.printL(5, "Constraints:")
        for const in constraints_list:
            Graph.log.printL(5, "\t%s:%s" % const)
        expand = 1

        node_map = NodeMap()

        nodes = set()
        edges_data = {}
        done_constraints = {}
        done_edges = []
        done_expand = []
        for next_elem, atts_elem in constraints_list:
            if Graph.isEG(next_elem):
                edges_data[next_elem] = {'constraints': atts_elem.items() ,'match_edges': None}
                edges_data[next_elem]['comp'] = Graph.onlyComp(edges_data[next_elem]['constraints'])
                nodes.update(next_elem)
            else:
                #### MAP NODE ALREADY
                expand = node_map.mapNode(next_elem, self.matchingElements(Graph.ND, atts_elem.items()))
                done_constraints[next_elem] = atts_elem

        for node in nodes - node_map.nodesSet():
            node_map.addNode(node)

        ### DONE WITH NODES, LET'S TAKE CARE OF EDGES NOW
        expand = 1
        frag_edges, closing_edges = node_map.edgesFragments(edges_data.keys())
        ### There should not be any closing edge since we are starting from scratch

        ### prepare connectivity infos
        heads_present = node_map.containedNodes(query_nodes)
        head_fragments = node_map.fragmentsNodes(heads_present)

        next_map = -1 ### just to get into the loop
        
        while next_map is not None and expand > 0:
            #### choose next edge to map
            next_map = None
            candidates_categories = [set(), set(), set()]
            for edge in set(edges_data.keys())-set(done_edges):
                if not edges_data[edge]['comp']:
                    inheads = 0
                    for endi in [0,1]:
                        if node_map.nodeFragment(edge[endi]) in head_fragments:
                            inheads += 1
                    candidates_categories[inheads].add(edge)

            while next_map is None and len(candidates_categories) > 0:
                edges_candidates = candidates_categories.pop()
                for edge in edges_candidates:
                    if edges_data[edge]['match_edges'] is None:
                        edges_data[edge]['match_edges'] = self.matchingElements(Graph.EG, edges_data[edge]['constraints'])

                    score = node_map.estimateCount(edge, edges_data[edge]['match_edges'])
                    if max_score > 0 and score > max_score:
                        Graph.log.printL(5, "TOOO HIGH SCORE BOTH %s, (%d)" % (edge, score))
                        score *= -1
                    if score > 0 and (next_map is None or next_map[1] > score):
                        next_map = (edge, score)

            if next_map is not None:
                next_elem = next_map[0]
                
                Graph.log.printL(5, "Now mapping %s, (%d) JOINING %d" % (next_elem, expand, next_map[1]))

                expand, new_fragid, old_fragid = node_map.mapJoiningEdge(next_elem, edges_data[next_elem]['match_edges'], oi_check)
                frag_edges[new_fragid].remove(next_elem)

                ### update infos about head-connectedness, frags cardinalities, etc.
                done_constraints[next_elem] = dict(edges_data[next_elem]['constraints'])
                done_edges.append(next_elem)
                done_expand.append(expand)
                head_fragments = node_map.fragmentsNodes(heads_present)

                if len(head_fragments) == 1 and node_map.sameFragment((next_elem[0], heads_present[0])):
                    details = {"constraints": done_constraints, "status": expand, "done": done_edges,
                               "head_fragment": node_map.nodeFragment(heads_present[0])}
                    yield node_map, details
                    
                ### Find closing edges and close...
                closing_edges = list(frag_edges[new_fragid] & frag_edges[old_fragid])
                frag_edges[new_fragid].update(frag_edges[old_fragid])
                while len(closing_edges) > 0 and expand > 0:
                    next_closing = closing_edges.pop()

                    ### next_elem has been remove from one side so it is not among closing edges
                    Graph.log.printL(5, "Now mapping %s, (%d) CLOSING" % (next_closing, expand))
                    cl_expand = node_map.mapClosingEdge(next_closing, edges_data[next_closing]['constraints'], self)
                    done_edges.append(next_closing)
                    done_expand.append(cl_expand)
                    if cl_expand < expand:
                        expand = cl_expand
                        done_constraints[next_closing] = dict(edges_data[next_closing]['constraints'])

                        if len(head_fragments) == 1 and node_map.sameFragment((next_closing[0], heads_present[0])):
                            details = {"constraints": done_constraints, "status": expand, "done": done_edges, 
                                        "head_fragment": node_map.nodeFragment(heads_present[0])}
                            yield node_map, details
        if len(set(edges_data.keys())-set(done_edges)) > 0:
            details = {"constraints": done_constraints, "status": -expand, "done": done_edges,
                       "head_fragment": node_map.nodeFragment(heads_present[0])}
            yield node_map, details
        

class Data:

    relevant = None

    def __init__(self, pred_types = None):
        self.graph = Graph()
        if pred_types is None:
            self.pred_types = {Graph.ND: None, Graph.EG: None} # edges, nodes
        else:
            self.pred_types = pred_types

    def infoStr(self):
        istr = ""
        if self.graph is not None:
            for tE in [Graph.ND, Graph.EG]:
                # istr += "Type %d #elements %d " % (tE, len(self.graph.elems[tE]))
                # istr += "%d attributes" % (len(self.graph.elems_pred[tE]))
                num = 0
                cat = 0
                vals = 0
                if self.pred_types[tE] is not None:
                    for att, typ in self.pred_types[tE].items():
                        if self.graph.elems_pred[tE].has_key(att):
                            if typ == 'cat':
                                vals += len(self.graph.elems_pred[tE][att])
                                cat += 1
                                # istr += ", %s:%d values" % (att, len(self.graph.elems_pred[tE][att]))
                            else:
                                num += 1
                                # istr += ", %s:numerical" % (att)
                istr += "#%d=%d %d num, %d cat (%d vals) " % (tE, len(self.graph.elems[tE]), num, cat, vals)
        return istr
    
    def copy(self):
        tmp = Data()
        tmp.graph = self.graph.copy()
        tmp.pred_types = copy.deepcopy(self.pred_types)
        return tmp

    def filteredPredTypes(self, pred_match=None):
        tmp = [{}, {}] # edges, nodes
        for tE in [Graph.EG, Graph.ND]:
            for predv, tP in self.pred_types[tE].items():
                if pred_match is None or not predv in pred_match:
                    tmp[tE][predv]=tP
        return tmp

    def parsePreds(self, fp):
        self.pred_types = [{}, {}]
        for line in fp:
            tmp = re.match("^##PRED (?P<tE>[01]) (?P<pred>[\w:.,;#=]+) (?P<tP>[a-z]+)", line)
            if tmp is not None:
                self.pred_types[int(tmp.group("tE"))][tmp.group("pred")]=tmp.group("tP")

    def toClausePreds(self):
        tmp_preds = {Graph.ND: {}, Graph.EG: dict(self.pred_types[Graph.EG])}
        for pred, tP in self.pred_types[Graph.ND].items():
            if tP == "num":
                tmp_preds[Graph.EG][pred] = "comp"
            else:
                tmp_preds[Graph.ND][pred] = tP
        return tmp_preds
        
    def writePreds(self, fp):
        for tE in [Graph.EG, Graph.ND]:
            for pred, tP in self.pred_types[tE].items():
                fp.write("##PRED %d %s %s\n" % (tE, pred, tP))

    def writeTriples(self, fp):
        self.graph.writeTriples(fp)

    def writeBMG(self, fp):
        self.graph.writeBMG(fp)

    def writeLogic(self, fp):
        self.graph.writeLogic(fp)

    def writeGraph(self, filename, write_preds=True):
        ext = os.path.splitext(filename)[1].lower()
        fp = open(filename, "w")
        if write_preds:
            self.writePreds(fp)
        if ext == ".bmg":
            self.writeBMG(fp)
        elif ext == ".triples":
            self.writeTriples(fp)
        elif ext == ".db":
            self.writeLogic(fp)
        fp.close()
        
    def loadGraph(self, filename):
        ext = os.path.splitext(filename)[1].lower()
        mtdRead = None
        if ext == ".bmg":
            mtdRead = readBMGLine
        elif ext == ".triples":
            mtdRead = readTripleLine
        elif ext == ".db":
            mtdRead = readLogicLine

        if mtdRead is not None:
            fp = open(filename)
            for line in fp:
                for triple in mtdRead(line):
                    self.graph.parseTriple(triple, self.pred_types)
            fp.close()
        
        if self.pred_types[Graph.ND] is None:
            self.pred_types[Graph.ND] = dict([(att, 'cat') for att in self.graph.elems_pred[Graph.ND].keys()])
        if self.pred_types[Graph.EG] is None:
            self.pred_types[Graph.EG] = dict([(att, 'cat') for att in self.graph.elems_pred[Graph.EG].keys()])
            
    def readGraph(self, filename):
        ext = os.path.splitext(filename)[1].lower()
        tmp = None
        mtdRead = None
        if ext == ".bmg":
            mtdRead = readBMGLine
        elif ext == ".triples":
            mtdRead = readTripleLine
        elif ext == ".db":
            mtdRead = readLogicLine

        if mtdRead is not None:
            fp = open(filename)
            tmp = Graph()
            for line in fp:
                for triple in mtdRead(line):
                    tmp.parseTriple(triple, self.pred_types)
            fp.close()
        return tmp

    def readGraphs(self, filename):
        ext = os.path.splitext(filename)[1].lower()
        graphs = {}
        labels = {}
        mtdRead = None
        if ext == ".bmg":
            mtdRead = readBMGLine
        elif ext == ".triples":
            mtdRead = readTripleLine
        elif ext == ".db":
            mtdRead = readLogicLine

        if mtdRead is not None:
            fp = open(filename)
            for line in fp:
                triples = mtdRead(line)
                if len(triples) == 0:
                    if len(line.split()) == 2:
                        tmp = re.match('^([A-Z]_)(?P<Label>\S+)\s+([A-Z]_)(?P<Key>[\w.,;#]+):(?P<Node>[\w.:,;#]+)\s*$', line)
                        if tmp is not None:
                            labels[tmp.group('Label')] = tmp.group("Key")
                for triple in triples:
                    key = None
                    new_triple = None
                    if triple is not None and len(triple) > 2:
                        new_triple = list(triple)
                        parts0 = triple[0].split(':', 1)
                        parts2 = triple[2].split(':', 1)
                        k0 = None; t0 = parts0[0];
                        k2 = None; t2 = parts2[0];
                        if len(parts0) > 1:
                            k0, t0 = parts0
                        if len(parts2) > 1:
                            k2, t2 = parts2
                            
                        if k2 is None or k0 == k2:
                            key = k0
                            new_triple[0] = t0
                            new_triple[2] = t2
                        if not graphs.has_key(key):
                            graphs[key] = Graph()
                        if key is None:
                            graphs[key].parseTriple(new_triple, {Graph.ND: {new_triple[1]:'cat'}, Graph.EG:{}})
                        else:
                            graphs[key].parseTriple(new_triple, self.pred_types)
            fp.close()
        return graphs, labels

class NodeMap:

    def __init__(self):
        self.mapped_nodes = {}
        self.fragments = []

    def instances(self):
        nodes = sorted(self.mapped_nodes.keys())
        inst = []
        for node in nodes:
            inst.append(self.mapped_nodes[node]['pl'])
        tmp = zip(*inst)
        return nodes, tmp

    def projectFragment(self, fragid):
        tmp = NodeMap()
        tmp.fragments = [[]]
        nodes = self.mapped_nodes.keys()
        for node in nodes:
            if self.mapped_nodes[node]['fragid'] == fragid:
                tmp.mapped_nodes[node] = copy.deepcopy(self.mapped_nodes[node])
                tmp.mapped_nodes[node]['fragid'] = 0
                tmp.fragments[0].append(node)
        return tmp

    def translateNode(self, node_from, node_to):
        if self.mapped_nodes.has_key(node_from) and not self.mapped_nodes.has_key(node_to):
            mp = self.mapped_nodes.pop(node_from)
            self.mapped_nodes[node_to] = mp

            self.fragments[self.mapped_nodes[node_to]['fragid']].remove(node_from)
            self.fragments[self.mapped_nodes[node_to]['fragid']].append(node_to)
            return True
        return False

    def containedNodes(self, nodes):
        if nodes is not None:
            return [q for q in nodes if self.mapped_nodes.has_key(q)]
        return []

    def fragmentsNodes(self, nodes):
        fnd = set()
        for node in nodes:
            fnd.add(self.mapped_nodes[node]['fragid'])
        return fnd

    def isMapped(self, node):
        return self.mapped_nodes[node]['pl'] is not None

    def hasNode(self, node):
        return self.mapped_nodes.has_key(node)

    def addNode(self, node):
        self.mapped_nodes[node] = {'fragid':len(self.fragments), 'pl': None, 'nd': None, 'mp': None}
        self.fragments.append([node])
                        
    def getFragments(self):
        if self.fragments is None:
            mf = max([data['fragid'] for data in self.mapped_nodes.values()])
            self.fragments = [[] for i in range(mf+1)]
            for node, data in self.mapped_nodes.items():
                self.fragments[data['fragid']].append(node)
        return self.fragments

    def sameFragment(self, edge):
        return self.mapped_nodes[edge[0]]['fragid'] == self.mapped_nodes[edge[1]]['fragid']

    def nodeFragment(self, node):
        return self.mapped_nodes[node]['fragid']

    def nodeMapsTo(self, node):
        if self.mapped_nodes[node]['mp'] is None:
            self.mapped_nodes[node]['mp'] = set(self.mapped_nodes[node]['pl'])
        return self.mapped_nodes[node]['mp']

    def nodeInstancesDict(self, node):
        if self.mapped_nodes[node]['nd'] is None:
            self.mapped_nodes[node]['nd'] = {}
            if self.mapped_nodes[node]['pl'] is not None:
                for ii, node_p in enumerate(self.mapped_nodes[node]['pl']):
                    self.mapped_nodes[node]['nd'].setdefault(node_p,[]).append(ii)
        return self.mapped_nodes[node]['nd']

    def nodeInstancesCounts(self, node):
        return dict([(node, len(inst)) for (node, inst) in self.nodeInstancesDict(node).items()])
    
    def copy(self):
        tmp = NodeMap()
        tmp.mapped_nodes = copy.deepcopy(self.mapped_nodes)
        tmp.fragments = copy.deepcopy(self.fragments)
        return tmp

    def nbMinMapsTo(self):
        nodes = self.nodesSet()
        if len(nodes) == 0:
            return 0
        else:
            return min([len(self.nodeMapsTo(node)) for node in nodes])

    def nbInstances(self, node):
        if self.mapped_nodes[node]['pl'] is not None:
            return len(self.mapped_nodes[node]['pl'])
        else:
            return 0

    def totInstances(self):
        return max([0]+[len(k['pl']) for k in self.mapped_nodes.values() if k['pl'] is not None])
            
    def nbNodes(self):
        return len(self.mapped_nodes)

    def nodesSet(self):
        return set(self.mapped_nodes.keys())
        
    def nodeInstance(self, node, pos):
        return self.mapped_nodes[node]['pl'][pos]
    
    def getSingletons(self):
        singletons = {}
        for node in self.mapped_nodes.keys():
            maps = self.nodeMapsTo(node)
            if len(maps) == 1:
                singletons[node] = list(maps)[0]
        return singletons

    def edgesFragments(self, edges):
        frag_edges = dict([(ii, set()) for ii in range(len(self.fragments))])
        closing_edges = []
        for edge in edges:
            for endi in [0,1]:
                frag_edges.setdefault(self.nodeFragment(edge[endi]), set()).add(edge)
            if self.sameFragment(edge):
                closing_edges.append(edge)

        return frag_edges, closing_edges


    def estimateCount(self, next_edge, match_edges):
        counts = [self.nodeInstancesCounts(next_edge[0]), self.nodeInstancesCounts(next_edge[1])]

        if self.isMapped(next_edge[0]) and self.isMapped(next_edge[1]):
            score = sum([counts[0].get(n[0], 0)*counts[1].get(n[1], 0) for n in match_edges]) 

        elif self.isMapped(next_edge[0]) or self.isMapped(next_edge[1]):
            if self.isMapped(next_edge[0]):
                mapped_side = 0
            else:
                mapped_side = 1                
            score = sum([counts[mapped_side].get(n[mapped_side], 0) for n in match_edges]) 

        else:
            score = len(match_edges)
        return score

    def mapNode(self, next_elem, match_nodes):
        if not self.hasNode(next_elem):
            self.addNode(next_elem)
 
        if self.mapped_nodes[next_elem]['pl'] is None:
            self.mapped_nodes[next_elem]['pl'] = match_nodes
            self.mapped_nodes[next_elem]['nd'] = None

        else:
            i = 0
            while i < len(self.mapped_nodes[next_elem]['pl']):
                if not self.mapped_nodes[next_elem]['pl'][i] in match_nodes:
                    for node_pl in self.fragments[self.mapped_nodes[next_elem]['fragid']]:
                        self.mapped_nodes[node_pl]['pl'].pop(i)
                else:
                    i+=1
            for node_pl in self.fragments[self.mapped_nodes[next_elem]['fragid']]:
                self.mapped_nodes[node_pl]['nd'] = None
        return len(self.mapped_nodes[next_elem]['pl'])
 
    def mapClosingEdge(self, next_edge, edge_constraints, graph):
        keep_positions = []

        for i in range(len(self.mapped_nodes[next_edge[0]]['pl'])):
            if graph.elementMatch((self.mapped_nodes[next_edge[0]]['pl'][i], self.mapped_nodes[next_edge[1]]['pl'][i]), edge_constraints):
                keep_positions.append(i)

        ### delete non matching ones, obviously there should at least be some...
        ##############
        expand = len(keep_positions)
        Graph.log.printL(7, "%d KEEP POSITIONS" % expand)

        if len(keep_positions) > 0 and len(keep_positions) < len(self.mapped_nodes[next_edge[0]]['pl']):
            ##print "#KEEP_POS", len(keep_positions), len(self.mapped_nodes[next_edge[0]]['pl'])
            for node_pl in self.fragments[self.mapped_nodes[next_edge[0]]['fragid']]:
                self.mapped_nodes[node_pl]['pl'] = [self.mapped_nodes[node_pl]['pl'][i] for i in keep_positions] 
                self.mapped_nodes[node_pl]['nd'] = None
        return expand
                
    def mapJoiningEdge(self, next_edge, match_edges, oi_check=True):
        new_fragid = self.mapped_nodes[next_edge[0]]['fragid']
        old_fragid = self.mapped_nodes[next_edge[1]]['fragid']    

        expand = len(match_edges)
        Graph.log.printL(7, "%d MATCHING EDGES" % expand)
        if expand <= 0:
            return expand, new_fragid, old_fragid

        if self.isMapped(next_edge[0]) and self.isMapped(next_edge[1]):
            expand = self.mapJoiningEdgeBoth(next_edge, match_edges, new_fragid, old_fragid, oi_check)

        elif self.isMapped(next_edge[0]) or self.isMapped(next_edge[1]):
            expand = self.mapJoiningEdgeOne(next_edge, match_edges, new_fragid, old_fragid, oi_check)

        else:
            expand = self.mapJoiningEdgeNeither(next_edge, match_edges, new_fragid, old_fragid)

        self.fragments[new_fragid] += self.fragments[old_fragid]
        self.fragments[old_fragid] = []
        return expand, new_fragid, old_fragid

    def mapJoiningEdgeNeither(self, next_edge, match_edges, new_fragid, old_fragid):
        Graph.log.printL(5, "NEITHER mapped...")
        ### should be two lonely nodes:
        ### len(self.fragments[new_fragi])*len(self.fragments[old_fragi]) == 1
        tmp = zip(*match_edges)
        expand = len(match_edges)

        (self.mapped_nodes[next_edge[0]]['pl'], self.mapped_nodes[next_edge[1]]['pl']) = (list(tmp[0]), list(tmp[1]))
        self.mapped_nodes[next_edge[1]]['fragid'] = new_fragid
        self.mapped_nodes[next_edge[0]]['nd'] = None
        self.mapped_nodes[next_edge[1]]['nd'] = None
        return expand

    def mapJoiningEdgeBoth(self, next_edge, match_edges, new_fragid, old_fragid, oi_check=True):
        Graph.log.printL(5, "BOTH mapped...")
        joining = [[],[]]

        node_instances = (self.nodeInstancesDict(next_edge[0]), self.nodeInstancesDict(next_edge[1]))
        ### for each matching edge, check for both ends whether there is a mapping to corresponding clause node
        ### if there is add the pair of instances lists to join
        ##############
        for i in range(len(match_edges)):
            for instance_Lid in node_instances[0].get(match_edges[i][0], []):
                for instance_Rid in node_instances[1].get(match_edges[i][1], []):
                    ### Object Identity: check that no two nodes in the new instance are mapped to same graph node
                    joini = True
                    if oi_check:
                        nodes_map = [(node, 0) for node in self.fragments[new_fragid]] + [(node, 1) for node in self.fragments[old_fragid]]
                        nb_nodes = len(nodes_map)
                        positions = (instance_Lid, instance_Rid)
                        joini = ( len(set( [self.mapped_nodes[node_pl]['pl'][positions[side]] for (node_pl, side) in nodes_map])) == nb_nodes)
                    if joini:
                        joining[0].append(instance_Lid)
                        joining[1].append(instance_Rid)

        expand = len(joining[0])
        Graph.log.printL(7, "BOTH RESULTS IN %s" % expand)

        ### join the list of instances of fragments
        for endi in [0,1]:
            nodes = self.fragments[self.mapped_nodes[next_edge[endi]]['fragid']]
            for node_pl in nodes:
                self.mapped_nodes[node_pl]['fragid'] = new_fragid
                self.mapped_nodes[node_pl]['pl'] = [self.mapped_nodes[node_pl]['pl'][position] for position in joining[endi]] 
                self.mapped_nodes[node_pl]['nd'] = None
        return expand

    def mapJoiningEdgeOne(self, next_edge, match_edges, new_fragid, old_fragid, oi_check=True):
        Graph.log.printL(5, "ONE mapped...")
        joining = []

        node_instances = (self.nodeInstancesDict(next_edge[0]), self.nodeInstancesDict(next_edge[1]))
        mapped_side = 0
        if len(node_instances[0]) == 0:
            mapped_side = 1

        self.mapped_nodes[next_edge[1-mapped_side]]['pl'] = []
        for i in range(len(match_edges)):
            for instance_Mid in node_instances[mapped_side].get(match_edges[i][mapped_side], []):
                ### Object Identity: check that no two nodes in the new instance are mapped to same graph node
                ##############
                if not oi_check or \
                       not (match_edges[i][1-mapped_side] in [self.mapped_nodes[node_pl]['pl'][instance_Mid] for node_pl in self.fragments[self.mapped_nodes[next_edge[mapped_side]]['fragid']] ]):
                    self.mapped_nodes[next_edge[1-mapped_side]]['pl'].append(match_edges[i][1-mapped_side])
                    joining.append(instance_Mid)

        expand = len(joining)
        Graph.log.printL(7, "ONE RESULTS IN %d" % expand)

        self.mapped_nodes[next_edge[1-mapped_side]]['fragid'] = new_fragid
        self.mapped_nodes[next_edge[1-mapped_side]]['nd'] = None
        nodes = self.fragments[self.mapped_nodes[next_edge[mapped_side]]['fragid']]
        for node_pl in nodes:
            self.mapped_nodes[node_pl]['fragid'] = new_fragid
            self.mapped_nodes[node_pl]['pl'] = [self.mapped_nodes[node_pl]['pl'][position] for position in joining]
            self.mapped_nodes[node_pl]['nd'] = None
        return expand


class TuplesCover:
    BLUE = 0
    RED = 1
    OUT = 2
    
    def __init__(self, query_nodes=None, nodes_map=None, blues=None, strict=False):
        self.sets = [set(), set(), set()] ## positive/negative/out cover
        self.tuples = {} ### maps tuples to the instances
        
        if query_nodes is not None and nodes_map is not None:
            self.computeTuples(query_nodes, nodes_map)

            if blues is not None:
                self.computeRedBlue(blues)
                if ( strict and blues != self.sets[TuplesCover.BLUE] ): print "Something badly wrong with mining/mapping..."
            else:
                self.computeRedBlue(set())

    def computeTuples(self, query_nodes, nodes_map):
        for j in range(nodes_map.totInstances()):
            current_tuple = tuple([nodes_map.nodeInstance(query_node,j) for query_node in query_nodes])
            if self.tuples.has_key(current_tuple):
                self.tuples[current_tuple].append(j)
            else:
                self.tuples[current_tuple] = [j]

    def computeRedBlue(self, blues):
        if Data.relevant is not None:
            all_tuples = set([ct for ct in self.tuples.keys() if (ct[0] in Data.relevant and ct[1] in Data.relevant)])
        else:
            all_tuples = set(self.tuples.keys())
        self.sets[TuplesCover.BLUE] = set(blues) & all_tuples
        self.sets[TuplesCover.RED] = all_tuples - self.sets[TuplesCover.BLUE]
        self.sets[TuplesCover.OUT] = set(blues) - all_tuples
                        
    def copy(self):
        tmp = TuplesCover()
        tmp.tuples = copy.deepcopy(self.tuples)
        tmp.sets = copy.deepcopy(self.sets)
        return tmp

    def resetRedBlue(self):
        self.sets = [set(), set(), set()]

    def blueTuples(self):
        return self.sets[TuplesCover.BLUE]
    def redTuples(self):
        return self.sets[TuplesCover.RED]
    def outTuples(self):
        return self.sets[TuplesCover.OUT]
    def tuplesParts(self, parts = []):
        supp = set()
        for b in parts:
            supp |= self.sets[b]
        return supp
    def support(self):
        return self.sets[TuplesCover.BLUE] | self.sets[TuplesCover.RED]
    def ordSets(self):
        return set([sorted(k) for k in self.support()])

    def nbTuplesParts(self, parts = []):
        return sum([len(self.sets[b]) for b in parts])
    def nbBlueTuples(self):
        return len(self.sets[TuplesCover.BLUE])
    def nbRedTuples(self):
        return len(self.sets[TuplesCover.RED])
    def nbOutTuples(self):
        return len(self.sets[TuplesCover.OUT])
    def nbSupport(self):
        return len(self.sets[TuplesCover.BLUE]) + len(self.sets[TuplesCover.RED])


    def functionalityIndices(self, typF="UNION"):
        listP = []
        if typF == "INTER":
            listP = self.sets[TuplesCover.BLUE]
        elif typF == "SUPP":
            listP = self.tuples.keys()
        elif typF == "UNION":
            listP = list(self.sets[TuplesCover.OUT]) + self.tuples.keys()
        elif typF == "OPP":
            listP = list(self.sets[TuplesCover.OUT]) + list(self.sets[TuplesCover.BLUE])

        nb_t = float(len(listP))
        return [len(set(tp))/nb_t for tp in zip(*listP)]


    def PCAData(self, sides=[0,1], typF="OPP", thres=0.5):
        nb_sides = []
        (v, pi) = (0, 0)
        if self.nbBlueTuples() > 0:
            (v,pi) = sorted([(v,i) for i,v in enumerate(self.functionalityIndices(typF))])[-1]
            if v > thres:
                nb_sides = [0,0]
                b_to = set([t[pi] for t in self.sets[TuplesCover.BLUE]])
                tokens = [self.sets[TuplesCover.OUT], self.sets[TuplesCover.RED]]
                for side in sides:
                    sset = set([t[pi] for t in tokens[1-side]])
                    sset |= b_to
                    nb_sides[side] = len([1 for s in tokens[side] if s[pi] in sset])
        return self.nbBlueTuples(), pi, v, nb_sides
    def stPCAPrecision(self, typF="OPP", thres=0.5):
        (nb, pi, v, nrs) = self.funcData([1], typF, thres)
        if len(nrs) == 2:
            return float(nb)/(nb+nrs[1])
        return 0
    def stPCARecall(self, typF="OPP", thres=0.5):
        (nb, pi, v, nrs) = self.PCSData([0], typF, thres)
        if len(nrs) == 2:
            return float(nb)/(nb+nrs[0])
        return 0
    def stPCATopPR(self, typF="OPP", thres=0.5):
        (nb, pi, v, nrs) = self.PCAData([0,1], typF, thres)
        if len(nrs) == 2:
            return float(nb)/(nb+min(nrs[0],nrs[1]))
        return 0
    def stPCAJaccard(self, typF="OPP", thres=0.5):
        (nb, pi, v, nrs) = self.PCAData([0,1], typF, thres)
        if len(nrs) == 2:
            return float(nb)/(nb+nrs[0]+nrs[1])
        return 0
    def PCAStr(self, typF="OPP", thres=0.5):
        (nb, pi, v, nrs) = self.PCAData([0,1], typF, thres)
        if len(nrs) == 2:
            return "source=%d func_ind=%f pca_red=%d pca_out=%d" % (pi, v, nrs[0], nrs[1])
        return "source=%d func_ind=%f" % (pi, v)

    def funcData(self, sides, typF="UNION", thres=0.5):
        nb_sides = []
        (v, pi) = (0, 0)
        if self.nbBlueTuples() > 0:
            (v,pi) = sorted([(v,i) for i,v in enumerate(self.functionalityIndices(typF))])[-1]
            if v > thres:
                nb_sides = [0, 0]
                blue_tokens = set([s[pi] for s in self.sets[TuplesCover.BLUE]])
                tokens = [self.sets[TuplesCover.OUT], self.sets[TuplesCover.RED]]
                for side in sides:
                    nb_sides[side] = len([1 for s in tokens[side] if s[pi] in blue_tokens])
        return self.nbBlueTuples(), pi, v, nb_sides
    def stFuncPrecision(self, typF="UNION", thres=0.5):
        (nb, pi, v, nrs) = self.funcData([1], typF, thres)
        if len(nrs) == 2:
            return float(nb)/(nb+nrs[1])
        return 0
    def stFuncRecall(self, typF="UNION", thres=0.5):
        (nb, pi, v, nrs) = self.funcData([0], typF, thres)
        if len(nrs) == 2:
            return float(nb)/(nb+nrs[0])
        return 0
    def stFuncTopPR(self, typF="UNION", thres=0.5):
        (nb, pi, v, nrs) = self.funcData([0,1], typF, thres)
        if len(nrs) == 2:
            return float(nb)/(nb+min(nrs[0],nrs[1]))
        return 0
    def stFuncJaccard(self, typF="UNION", thres=0.5):
        (nb, pi, v, nrs) = self.funcData([0,1], typF, thres)
        if len(nrs) == 2:
            return float(nb)/(nb+nrs[0]+nrs[1])
        return 0
    def funcStr(self, typF="UNION", thres=0.5):
        (nb, pi, v, nrs) = self.funcData([0,1], typF, thres)
        if len(nrs) == 2:
            return "source=%d func_ind=%f func_red=%d func_out=%d" % (pi, v, nrs[0], nrs[1])
        return "source=%d func_ind=%f" % (pi, v)


    def stDiffRB(self):
        return self.nbBlueTuples()-self.nbRedTuples()
    def stRatioRB(self):
        btp = float(self.nbBlueTuples())
        if btp > 0:
            return self.nbRedTuples()/btp
        else:
            return float('Inf')
    def stPrecision(self):
        btp = float(self.nbBlueTuples())
        if btp > 0:
            return btp/(btp + self.nbRedTuples())
        else:
            return 0
    def stRecall(self):
        btp = float(self.nbBlueTuples())
        if btp > 0:
            return btp/(btp + self.nbOutTuples())
        else:
            return 0
    def stJaccard(self):
        btp = float(self.nbBlueTuples())
        if btp > 0:
            return btp/(btp + self.nbRedTuples() + self.nbOutTuples())
        else:
            return 0
    def stTopPR(self):
        btp = float(self.nbBlueTuples())
        if btp > 0:
            return btp/(btp + min(self.nbRedTuples(), self.nbOutTuples()))
        else:
            return 0

    def stEitherTop(self, typF="OPP"):
        if self.nbBlueTuples() > 0:
            tPR = self.stTopPR()
            if typF == "FUNC":
                tO = self.stFuncTopPR()
            else:
                tO = self.stPCATopPR(typF)
            return max(tPR, tO)
        else:
            return 0


class Clause:

    #basic_okey = [('ratio', 1), ('support', -1), ('nb_constraints', 1), ('nb_instances', 1)]
    basic_okey = [('jaccard', -1), ('support', -1), ('nb_nodes', 1), ('nb_edges', 1)]
    elems_check = {'precision': {'method': 'self.cover.stPrecision()', 'close':0},
                   'recall': {'method': 'self.cover.stRecall()', 'close':0},
                   'jaccard':  {'method': 'self.cover.stJaccard()', 'close':0},
                   'top': {'method': 'self.cover.stTopPR()', 'close':0},
                   'func_precision': {'method': 'self.cover.stFuncPrecision()', 'close':0},
                   'func_recall': {'method': 'self.cover.stFuncRecall()', 'close':0},
                   'func_jaccard':  {'method': 'self.cover.stFuncJaccard()', 'close':0},
                   'func_top': {'method': 'self.cover.stFuncTopPR()', 'close':0},
                   'pca_precision': {'method': 'self.cover.stPCAPrecision()', 'close':0},
                   'pca_recall': {'method': 'self.cover.stPCARecall()', 'close':0},
                   'pca_jaccard':  {'method': 'self.cover.stPCAJaccard()', 'close':0},
                   'pca_top': {'method': 'self.cover.stPCATopPR()', 'close':0},
                   'either_top': {'method': 'self.cover.stEitherTop()', 'close':0},
                   'examples':  {'method': 'self.cover.nbBlueTuples()', 'close':-1},
                   'support':  {'method': 'self.cover.nbSupport()', 'close':-1},
                   'nb_blue':  {'method': 'self.cover.nbBlueTuples()', 'close':-1},
                   'nb_red':  {'method': 'self.cover.nbRedTuples()', 'close':-1},
                   'nb_out':  {'method': 'self.cover.nbOutTuples()', 'close':1},
                   'ratio':  {'method': 'self.cover.stRatioRB()', 'close':0},
                   'diff':  {'method': 'self.cover.stDiffRB()', 'close':0},
                   'inter':  {'method': 'self.cover.nbBlueTuples()', 'close':-1},
                   'nb_nodes':  {'method': 'self.constraints.nbNodes()', 'close':1},
                   'nb_edges':  {'method': 'self.constraints.nbEdges()', 'close':1},
                   'nb_constraints':  {'method': 'self.constraints.nbConstraints()', 'close':1},
                   'nb_instances':  {'method': 'self.nodes_map.totInstances()', 'close':0},
                   'maps':  {'method': 'self.nodes_map.nbMinMapsTo()', 'close':-1}}
    
    def __init__(self, nid='TMP', constraints=None, nm=None, query_nodes=None, edges=set(), data_preds=None):
        self.id = nid
        self.nb_org = 0
        if query_nodes is not None:
            self.query_nodes = copy.deepcopy(query_nodes)
        else:
            self.query_nodes = None

        self.constraints = None
        if isinstance(constraints, Graph):
            self.constraints = constraints
        elif type(constraints) in [tuple, list]:
            self.constraints = Graph()
            if data_preds is None:
                data_preds = {Graph.ND:None, Graph.EG:None}
            self.constraints.parseTriples(constraints, data_preds)

        ### in case we provide graph data
        self.nodes_map = None
        if isinstance(nm, Graph):
            if self.constraints is not None:
                constraints_list = self.constraints.getConstraintsList()
                mapped_nodes, details = nm.mapClause(constraints_list, self.query_nodes)
                if details['status'] > 0:
                    self.nodes_map = mapped_nodes

        ### else if we provide readily mapped nodes
        #### WARNING! NO COPY
        elif isinstance(nm, NodeMap):
            self.nodes_map = nm

        self.cover = None
        if self.nodes_map is not None and self.query_nodes is not None:
            self.cover = TuplesCover(self.query_nodes, self.nodes_map, edges)

    def instancesNodes(self):
        return self.nodes_map.instances()

    def instancesQTDict(self):
        intermediates = list(self.nodes_map.nodesSet() - set(self.query_nodes))
        inter_inst = dict([(n, {'instances': {}, 'maps_to': None}) for n in intermediates])
        for node in intermediates:
            inter_inst[node]['maps_to'] = self.nodes_map.nodeMapsTo(node)
            for qt, instances in self.cover.tuples.items():
                inter_inst[node]['instances'][qt] = set()
                for inst in instances:
                    inter_inst[node]['instances'][qt].add(self.nodes_map.nodeInstance(node, inst))
        return inter_inst

    def getCrit(self, crit, default=None):
        if Clause.elems_check.has_key(crit):
            return eval(Clause.elems_check[crit]['method'])
        return default

    def orderKey(self, parameters={}):
        if parameters.has_key('order_key'):
            methodOrder = eval("self.orderKeyM"+parameters['order_key'])
            if callable(methodOrder):
                return methodOrder(parameters)
        return self.orderKeyMBasic(parameters)

    def orderKeyGeneric(self, list_okeys):
        okey = []
        for crit, fact in list_okeys:
            okey.append(fact*self.getCrit(crit))
        return tuple(okey)

    def orderKeyMBasic(self, parameters):
        if parameters.has_key('list_okey'):
            list_okey = parameters['list_okey']
        else:
            list_okey=Clause.basic_okey
        return self.orderKeyGeneric(list_okey)

    def checkAccept(self, parameters):
        if parameters.has_key('check_accept'):
            methodCheck = eval("self.checkMAccept"+parameters['check_accept'])
            if callable(methodCheck):
                return methodCheck(parameters)
        return self.checkAcceptMBasic(parameters)

    def checkAcceptGeneric(self, list_checks):
        accept = True
        i = 0
        log_check = []
        while accept and i < len(list_checks):
            crit, thres, fact = list_checks[i]
            val = self.getCrit(crit, 0)
            accept &= (fact*val) >= (fact*thres)
            log_check.append((crit, val, accept))
            i+=1
        return accept, log_check

    def checkAcceptMBasic(self, parameters):
        if parameters.has_key('list_accept'):
            list_accept = parameters['list_accept']
        else:
            list_accept=[]
        return self.checkAcceptGeneric(list_accept)

    def checkExpand(self, parameters):
        if parameters.has_key('list_accept'):
            list_accept = parameters['list_accept']
        else:
            list_accept=[]
        return self.checkExpandGeneric(list_accept)

    def checkExpandGeneric(self, list_checks):
        accept = True
        i = 0
        log_check = []
        while accept and i < len(list_checks):
            crit, thres, fact = list_checks[i]
            if Clause.elems_check.has_key(crit) and fact*Clause.elems_check[crit]['close'] < 0:        
                val = self.getCrit(crit, 0)
                accept &= (fact*val) >= (fact*thres)
                log_check.append((crit, val, accept))
            i+=1
        return accept, log_check

    def getTypesPred(self):
        if self.constraints is not None:
            return self.constraints.getTypesPred()
        return []
        
    def labelSingletonMaps(self):
        for node, label in self.nodes_map.getSingletons().items():
            self.constraints.parseElem(Graph.ND, node, Graph.singleton_label, label)

    def hasSingletonMap(self):
        return len(self.nodes_map.getSingletons()) > 0

    def copy(self):
        tmp = Clause()
        tmp.id = 'copyOf' + self.getId()
        if self.constraints is not None:
            tmp.constraints = self.constraints.copy()
        if self.query_nodes is not None:
            tmp.query_nodes = copy.deepcopy(self.query_nodes)
        if self.nodes_map is not None:
            tmp.nodes_map = self.nodes_map.copy()
        if self.cover is not None:
            tmp.cover = self.cover.copy()
        return tmp

    def translate(self, translation_dict):
        done = []
        for node_from, node_to in translation_dict.items():
            if self.constraints.translateNode(node_from, node_to, False):
                self.nodes_map.translateNode(node_from, node_to)
                done.append(node_from)
        if len(done) > 0:
            self.constraints.updateElemsPreds()
        return done


    def dispShort(self):
        if self.cover is not None:
            return "Clause %s (%d/%d)" % (self.getId(), self.cover.nbBlueTuples(), self.cover.nbRedTuples())
        return "Clause %s (?)" % (self.getId())
        
    def getId(self):
        return str(self.id)
    
    def setId(self, nid):
        self.id = nid

    def support(self):
        if self.cover is not None:
            return self.cover.support()

    def ordSupport(self, order=None):
        tmp = []
        if self.cover is not None:
            if order is None:
                tmp = list(self.cover.support())
            elif order == -1:
                for tupleT in self.cover.support():
                    tmp.append(tuple(sorted(tupleT)))
            else:
                for tupleT in self.cover.support():
                    tmp.append(tuple([tupleT[o] for o in order]))
        tmp.sort()
        return tmp

    def supportKey(self, graph):
        sums = []
        tmtt = self.cover.support()

        for qi, qrn in enumerate(self.query_nodes):
            sums.append(sum((graph.nodeId(tupleT[qi]) for tupleT in tmtt)))

        # sums = [0  for qrn in self.query_nodes]
        # for tupleT in self.cover.support():
        #     for qi, qrn in enumerate(self.query_nodes):
        #         sums[qi] += graph.nodeId(tupleT[qi])
        order = sorted(range(len(sums)), key=lambda x: sums[x])
        # if len(sums) != len(set(sums)) and self.ordSupport(order) != self.ordSupport(order[-1::-1]):
        #     pdb.set_trace()
        #     print self.getId()

        return self.ordSupport(order)

    def supportKeyStr(self, graph):
        return  '+'.join(["%s-%s" % edge for edge in self.supportKey(graph)])

    def nbBlueTuples(self):
        if self.cover is not None:
            return len(self.cover.blueTuples())

    def nbRedTuples(self):
        if self.cover is not None:
            return len(self.cover.redTuples())

    def blueTuples(self):
        if self.cover is not None:
            return self.cover.blueTuples()

    def redTuples(self):
        if self.cover is not None:
            return self.cover.redTuples()

    def printBMG(self):
        return '\n'.join(sorted(self.listBMG()))

    def listBMG(self, replace_match='^', replace_subs='T_', color_str=None, with_info=[]):
        crit_info = []
        if with_info is not None:
            crit_info = set(['nb_blue','nb_red', 'nb_out'] + list(with_info))
                ### GRAPH OUTPUT
        if color_str is None:
            if self.cover is not None and with_info:
                color_str = "fill=%d/%d/255 " % (200*(1-self.cover.stPrecision()), 200*(1-self.cover.stRecall()))
            else:
                color_str = ""

        clause_graph = self.constraints.listBMG(replace_match, replace_subs)
        if self.cover is not None and len(crit_info) > 0:
            info_str = "# _attributes T_%s" % self.getId()
            fun = False; pca = False
            for crit in crit_info:
                if re.match("^func_", crit):
                    if not fun:
                        info_str += " " + self.cover.funcStr()
                        fun = True
                if re.match("^pca_", crit) or crit == 'either_top':
                    if not pca:
                        info_str += " " + self.cover.PCAStr()
                        pca = True
                else:
                    info_str += " %s=%s" % (crit, self.getCrit(crit, 0))
            clause_graph.append(info_str)

        if len(self.query_nodes) > 0 and with_info: ### SHOULD ALWAYS BE THE CASE
            node_anchor = self.query_nodes[0]
            tmp_lab = re.sub(replace_match, replace_subs, node_anchor)
            clause_graph.append("T_%s %s" % (self.getId(), tmp_lab))
            
        for node in self.constraints.nodesSet():
            tmp_lab = re.sub(replace_match, replace_subs, node)
            str_atts = color_str
            if self.nodes_map:
                str_atts += "nb_instances=%s nb_mapped=%s " % (self.nodes_map.totInstances(), len(self.nodes_map.nodeMapsTo(node)) )
            clause_graph.append("# _attributes %s %s"  % (tmp_lab, str_atts))
        return clause_graph



